
# `constraints` is passed in an iterable of (lower,upper)=>constr_fun(p).
#
# TODO (cg 2022/07/10 17:57): can this function be called without loading in
# SparseDiffTools.jl? I wonder if I need to do some kind of Val-based dispatch
# if not to make sure that those other branches only get hit if you actually
# provide a sparsity structure.
#
# TODO (cg 2022/07/12 12:05): Break out a different function that prepares the
# necessary functions, and then write another version of ipopt_optimize that
# accepts all the functions? I don't see much of a point in that because that
# doesn't actually have all that much boilerplate anymore. On the other hand,
# there are a lot of ifs all over this now. I feel like breaking this up to
# handle sparsity and mixed-mode AD would make this code much more readable.
function ipopt_optimize(obj, ini, constraints=(); 
                        jac_sparsity::SPBOOL_OR_NOTHING=nothing, 
                        mixed_ad=false, return_internals=false, 
                        box_lower=-1e22, box_upper=1e22, kwargs...)
  # check if there are constraints or not:
  hascons = !isempty(constraints)
  densej  = isnothing(jac_sparsity)
  # bring in some default arguments:
  optargs = merge(DEF_IPOPT_KWARGS, Dict(kwargs))
  # get the dimension of the problem:
  arglen  = length(ini)
  nconstr = hascons ? length(constraints) : 0
  # color vector for the jacobian if it is sparse:
  jcolorv = densej ? nothing : matrix_colors(jac_sparsity) 
  # set up box constraints for the args and, if applicable, the constraints:
  box_l = prepare_bounds(arglen, box_lower)
  box_u = prepare_bounds(arglen, box_upper)
  con_l = hascons ? [x.first[1] for x in constraints] : Float64[]
  con_u = hascons ? [x.first[2] for x in constraints] : Float64[]
  # function for the constraints:
  constrfuns  = hascons ? ntuple(j->constraints[j].second, nconstr) : ()
  constr_buf  = zeros(nconstr)
  constr_eval = (p,store) -> begin 
    for j in eachindex(constrfuns)
      @inbounds store[j] = constrfuns[j](p)
    end
    nothing
  end
  # If mixed AD is requested, create that object:
  if mixed_ad
    obj_wrapper = FoRWrapper(obj, ini)
  end
  # objective gradient and constraint jacobian:
  if mixed_ad
    objg = (p,g)->begin
      grad = obj_wrapper(:gradient, p)
      g .= grad
    end
  else
    objg = (p,g)->ForwardDiff.gradient!(g, obj, p)
  end
  jac_buf = densej ? zeros(nconstr, arglen) : Float64.(jac_sparsity)
  jac_nz  = densej ? arglen*nconstr : sum(!iszero, jac_sparsity)
  constr_eval_rev = (store, p) -> constr_eval(p, store)
  if densej 
    constr_jac = (p,r,c,v) -> begin 
      isnothing(v) && return jac_structure!(r, c, nconstr, length(p)) 
      ForwardDiff.jacobian!(jac_buf, constr_eval_rev, constr_buf, p) 
      v .= vec(jac_buf') 
    end
   else
    constr_jac = (p,r,c,v) -> begin 
      isnothing(v) && return jac_structure_sparse!(jac_sparsity, r, c)
      forwarddiff_color_jacobian!(jac_buf, constr_eval_rev, p, colorvec=jcolorv)
      populate_sparse_jacobian!(v, jac_sparsity, jac_buf)
    end
  end
  # Lagrangian hessian:
  lagrange = (p,l,o) -> begin
    out = obj(p)*o
    hascons || return out
    for j in eachindex(constrfuns)
      out += l[j]*constrfuns[j](p)
    end
    out
  end
  lagrange_h = (p,r,c,o,l,v) -> begin 
    isnothing(v) && return hes_structure!(r,c,length(p)) 
    if mixed_ad
      lagrange_wrapper = FoRWrapper(_p->lagrange(_p, l, o), p)
      lhess = lagrange_wrapper(:hessian, p)
    else
      lhess = ForwardDiff.hessian(_p->lagrange(_p, l, o), p)
    end
    linear_buffer_ltri_colmajor!(v, lhess) 
    nothing
  end
  # For debugging:
  return_internals && return (;obj, constr_eval, objg, constr_jac, lagrange_h)
  # Create the final problem:
  prob = CreateIpoptProblem(arglen,  box_l, box_u,     # arg count and box
                            nconstr, con_l, con_u,     # constr. count and box
                            jac_nz,                    # size of constr. jacobian
                            div(arglen*(arglen+1), 2), # size of hessian ltri
                            obj,                       # objective
                            constr_eval,               # constraint function
                            objg,                      # objective gradient
                            constr_jac,                # constraint jacobian
                            lagrange_h)                # lagrangian hessian
  # Add all the options:
  for (opt, val) in optargs
    ipopt_addoption!(prob, opt, val)
  end
  # Solve and return:
  prob.x = copy(ini)
  status = IpoptSolve(prob)
  (status=status, minimizer=copy(prob.x), minval=obj(prob.x))
end

# TODO (cg 2022/08/01 21:37): offer an interface for fn in-place.
function ipopt_nlsolve(fn, ini; box_lower=-1e22, box_upper=1e22, kwargs...)
  # bring in some default arguments:
  optargs = merge(DEF_IPOPT_KWARGS, Dict(kwargs))
  # get the dimension of the problem:
  arglen  = length(ini)
  nconstr = arglen
  # set up box constraints for the args and, if applicable, the constraints:
  box_l = prepare_bounds(arglen, box_lower)
  box_u = prepare_bounds(arglen, box_upper)
  con_l = zeros(length(ini))
  con_u = zeros(length(ini))
  # function for the constraints:
  constr_eval = (p,store) -> (store .= fn(p))
  # "constraint" jacobian:
  constr_jac = (p,r,c,v) -> begin 
    isnothing(v) && return jac_structure!(r, c, nconstr, nconstr) 
    v .= vec(ForwardDiff.jacobian(fn, p)')
  end
  # Lagrangian hessian:
  lagrange_h = (p,r,c,o,l,v) -> begin 
    isnothing(v) && return hes_structure!(r,c,length(p)) 
    fill!(v, 0.0)
    nothing
  end
  # Create the final problem:
  prob = CreateIpoptProblem(arglen,  box_l, box_u,        # arg count and box
                            nconstr, con_l, con_u,        # constr. count and box
                            arglen^2,                     # size of constr. jacobian
                            div(arglen*(arglen+1), 2),    # size of hessian ltri
                            _p->0.0,                      # objective
                            constr_eval,                  # constraint function
                            (p,g)->(g.=zeros(length(p))), # objective gradient
                            constr_jac,                   # constraint jacobian
                            lagrange_h)                   # lagrangian hessian
  # Add all the options:
  for (opt, val) in optargs
    ipopt_addoption!(prob, opt, val)
  end
  # Solve and return:
  prob.x = copy(ini)
  status = IpoptSolve(prob)
  (status=status, minimizer=copy(prob.x))
end

